// This software including the source code and documentation (the "SOFTWARE")
// is copyrighted.  You are free to use the SOFTWARE for any non-commercial
// purpose whatsoever provided you do not remove this copyright notice.
// If you wish to use the SOFTWARE for commercial purposes, please contact us
// to obtain a license.  The SOFTWARE is provided "as is", with absolutely no
// warranty.  Use it at your own risk.
// Website: http://www.math.ie/checker.html

#ifndef _graph_H
#define _graph_H  1

#include <stdio.h>
#include <vector>
#include <algorithm>

using namespace std;

// cheap integer array class
class IntArray
   {
   public:
      IntArray( int maxsize ) : _size( 0 ), _maxsize( maxsize ),
                                _data( new int[ _maxsize ] ) {}
      ~IntArray() { delete[] _data; }

      int& operator[]( int i ) { return _data[ i ]; }
      int operator[]( int i ) const { return _data[ i ]; }
      int GetSize() const { return _size; }
      void SetSize( int newsize ) { _size = newsize; }
      void PushBack( int val ) { _data[ _size++ ] = val; }

      void operator=( IntArray const& rhs )
         {
         delete[] _data;
         _maxsize = rhs._maxsize;
         _data    = new int[ _maxsize ];
         _size    = rhs._size;
         memcpy( _data, rhs._data, _size * sizeof( int ) );
         }

      void Print() const
         {
         for ( int i = 0;  i < _size;  i++ )
            printf( "%d ", _data[ i ] );
         printf( "\n" );
         }

   private:
      int  _size;
      int  _maxsize;
      int* _data;
   };


// an undirected graph
class Graph
   {
   public:
      // vertices are numbered 0, 1, ..., cVertices - 1
      Graph( int cVertices ) : m_cVertices( cVertices ),
                               m_ppEdges( new bool*[ cVertices ] )
         {
         for ( int i = 0;  i < m_cVertices;  i++ )
            {
            m_ppEdges[ i ] = new bool[ m_cVertices ];
            // set all entries to 'false' (no edges yet)
            memset( m_ppEdges[ i ], 0, sizeof( bool ) * m_cVertices );
            }
         }

      // bogus assignment operator to avoid compiler warning (default assignment
      // operator cannot be generated because some member variables are constant)
      void operator=( Graph const& ) { printf( "Not allowed." ); }

      ~Graph()
         {
         for ( int i = 0;  i < m_cVertices;  i++ )
            delete[] m_ppEdges[ i ];
         delete[] m_ppEdges;
         }

      // Does the edge (i,j) exist?
      bool AreAdjacent( int i, int j ) const
         {
         return m_ppEdges[ i ][ j ];
         }

      int VertexDegree( int i ) const
         {
         int nDeg = 0;
         for ( int j = 0;  j < m_cVertices;  j++ )
            {
            if ( AreAdjacent( i, j ) )
               nDeg++;
            }
         return nDeg;
         }

      // add an edge between the vertices i and j
      void AddEdge( int i, int j )
         {
         m_ppEdges[ i ][ j ] = m_ppEdges[ j ][ i ] = true;
         }

      void FindMaxClique( vector< int >* pvcClique )
         {
         // first order vertices by degree
         vector< int > vcVertices;
         for ( int i = 0;  i < m_cVertices;  i++ )
            vcVertices.push_back( i );
         int* degrees = new int[ m_cVertices ];
         for ( int i = 0;  i < m_cVertices;  i++ )
            degrees[ i ] = VertexDegree( i );

         IntArray vertices2( m_cVertices );
         while ( vcVertices.size() > 0 )
            {
            int nVertex = vcVertices[ 0 ];
            int nMinDeg = degrees[ nVertex ];
            // find the vertex of smallest degree in the graph consisting
            // of the vertices in vcVertices
            for ( unsigned i = 1;  i < vcVertices.size();  i++ )
               {
               if ( degrees[ vcVertices[ i ] ] < nMinDeg )
                  {
                  nVertex = vcVertices[ i ];
                  nMinDeg = degrees[ nVertex ];
                  }
               }
            vertices2.PushBack( nVertex );
            // delete this vertex from the graph
            vcVertices.erase( find( vcVertices.begin(), vcVertices.end(), nVertex ) );
            // correct degrees of remaining vertices
            for ( unsigned i = 0;  i < vcVertices.size();  i++ )
               {
               if ( AreAdjacent( nVertex, vcVertices[ i ] ) )
                  degrees[ vcVertices[ i ] ]--;
               }
            }
         delete[] degrees;
         m_pMaxClique = new IntArray( m_cVertices );
         IntArray clique( m_cVertices );
         FindCliqueInSubgraph( 1, 1, vertices2, clique );
         pvcClique->clear();
         for ( int i = 0;  i < m_pMaxClique->GetSize();  i++ )
            pvcClique->push_back( (*m_pMaxClique)[ i ] );
         delete m_pMaxClique;
         }

   private:
      // recursively find a maximal clique in the subgraph whose
      // vertices are listed in 'vertices'
      int FindCliqueInSubgraph( int iRound, int nMinSize, IntArray const& vertices, IntArray& clique )
         {
         if ( vertices.GetSize() <= 1 )
            {
            if ( clique.GetSize() + vertices.GetSize() > m_pMaxClique->GetSize() )
               {
               if ( vertices.GetSize() == 1 )
                  {
                  clique[ iRound - 1 ] = vertices[ 0 ];
                  clique.SetSize( iRound );
                  }
               *m_pMaxClique = clique;
               }

            return vertices.GetSize();
            }

         // 'nMinSize' gives the minimum clique size we are interested in
         // (note: nMinSize changes during the execution of the loop)
         int nMax = 0;
         IntArray vertices2( vertices.GetSize() - 1 );
         for ( int i = 0;  i < vertices.GetSize() - (nMinSize - 1);  i++ )
            {
            // add vertex vertices[ i ]
            clique[ iRound - 1 ] = vertices[ i ];
            clique.SetSize( iRound );
            int  cNonAdjacent = 0;
            bool bDoSearchSubGraph = true;
            // now find subgraph of all vertices adjacent to the vertex we just added
            vertices2.SetSize( 0 );
            for ( int j = i + 1;  j < vertices.GetSize();  j++ )
               {
               if ( AreAdjacent( vertices[ i ], vertices[ j ] ) )
                  vertices2.PushBack( vertices[ j ] );
               else
                  {
                  cNonAdjacent++;
                  if ( vertices.GetSize() - (i + 1) - cNonAdjacent < nMinSize - 1 )
                     {
                     bDoSearchSubGraph = false;
                     break;
                     }
                  }
               }

            if ( bDoSearchSubGraph )
               {
               int nSizeMaxCliqueSubGraph =
                  FindCliqueInSubgraph( iRound + 1,
                                        nMinSize > 1 ? (nMinSize - 1) : 1,
                                        vertices2, clique );
               if ( nSizeMaxCliqueSubGraph + 1 >= nMinSize )
                  {
                  nMax = nSizeMaxCliqueSubGraph + 1;
                  nMinSize = nMax + 1;
                  }
               }
            }

         return nMax;
         }

   private:
      int const      m_cVertices;
      bool** const   m_ppEdges;
      IntArray*      m_pMaxClique;
   };

#endif
