// This software including the source code and documentation (the "SOFTWARE")
// is copyrighted.  You are free to use the SOFTWARE for any non-commercial
// purpose whatsoever provided you do not remove this copyright notice.
// If you wish to use the SOFTWARE for commercial purposes, please contact us
// to obtain a license.  The SOFTWARE is provided "as is", with absolutely no
// warranty.  Use it at your own risk.
// Website: http://www.math.ie/checker.html

// checker
// searches a Sudoku grid for an n-clue puzzle
// to learn the syntax, just run the program with no parameters

#include <time.h>
#include <ctype.h>
#include <vector>
#include <iostream>
#include "unique.h"
#include "config.h"
#include "misc.h"
#include "mywin32.h"
#include "checker.h"
using namespace std;

void PrintSyntax();

int main( int argc, char const* argv[] )
   {
   if ( argc < 3 )
      {
      PrintSyntax();
      return 0;
      }

   int cClues = atoi( argv[ 2 ] );
   if ( cClues < 7  ||  cClues > 30 )
      {
      printf( "Number of clues should be at least 7 but not bigger than 30.\n" );
      return 0;
      }

   // parse command-line arguments
   bool bRandom = false;
   EnableFindAllUnavoidables( true );
   char const* pszFileExtraUnav = NULL;
   int cFirstClues = 0;
   int nPercentage = 100;
   for ( int i = 3;  i < argc;  i++ )
      {
      if ( 0 == strcmp( argv[ i ], "-solve" ) )
         {
         if ( ++i == argc )
            {
            // need one further argument
            printf( "Error: must specify percentage of puzzles to solve (0-100).\n" );          
            return -1;
            }
         else
            {
            nPercentage = atoi( argv[ i ] );
            if ( ! isdigit( argv[ i ][ 0 ] )  ||  nPercentage < 0  ||  nPercentage > 100 )
               {
               printf( "Error: percentage must be in range 0-100.\n" );
               return -1;
               }
            }
         }
      else if ( 0 == strcmp( argv[ i ], "-unav" ) )
         {
         if ( ++i == argc )
            {
            // need one further argument
            printf( "Error: must specify name of file containing external unavoidables.\n" );            
            return -1;
            }
         else
            pszFileExtraUnav = argv[ i ];
         }
      else if ( 0 == strcmp( argv[ i ], "-firstclues" ) )
         {
         if ( ++i == argc )
            {
            // need one further argument
            printf( "Error: must specify number of clues to fix.\n" );           
            return -1;
            }
         else
            cFirstClues = atoi( argv[ i ] );
         }
      else if ( 0 == strcmp( argv[ i ], "-x" ) )
         {
         if ( IsSudokuDG() )
            {
            printf( "Error: cannot use both -x and -dg simultaneously.\n" );           
            return -1;
            }
         EnableSudokuX();
         }
      else if ( 0 == strcmp( argv[ i ], "-dg" ) )
         {
         if ( IsSudokuX() )
            {
            printf( "Error: cannot use both -x and -dg simultaneously.\n" );           
            return -1;
            }
         EnableSudokuDG();
         }
      else if ( 0 == strcmp( argv[ i ], "-random" ) )
         {
         bRandom = true;
         }
      else if ( 0 == strcmp( argv[ i ], "-pseudos" ) )
         {
         EnableIncludePseudos();
         }
      else if ( 0 == strcmp( argv[ i ], "-uquick" ) )
         {
         // find only a limited collection of unavoidables (which is faster)
         EnableFindAllUnavoidables( false );
         }
      else if ( 0 == strcmp( argv[ i ], "-stats" ) )
         {
         // print some statistics during search
         EnableStats();
         }
      else
         printf( "Ignoring unknown command-line argument %s.\n", argv[ i ] );
      }

#ifdef _WIN32
   // assign lowest priority so that checker runs in the background and
   // does not slow down other processes
   SetPriorityClass( GetCurrentProcess(), IDLE_PRIORITY_CLASS );
#endif

   // initialise solver
   InitSolver();
   // read a grid from a file and compute all unavoidable sets
   Grid grid( argv[ 1 ], pszFileExtraUnav );
   // extract filename for name of puzzle file
   char* const psz = ToLowerCase( strdup( argv[ 1 ] ) );
   // chop the trailing '.txt' or '.dat'
   if ( strlen( psz ) > 4  &&
        (0 == strcmp( psz + strlen( psz ) - 4, ".txt" )  ||
         0 == strcmp( psz + strlen( psz ) - 4, ".dat" )) )
      {
      psz[ strlen( psz ) - 4 ] = '\0';
      }
   ClueChecker checker( grid, psz, cClues, nPercentage );
   if ( ! bRandom )
      checker.DoChecking( cFirstClues );
   else
      checker.DoRandomChecking();
   return 0;
   }


void PrintSyntax()
   {
   cout << PROG_EXECUTABLE_NAME " <gridfile> <n> [-pseudos] [-solve <percentage>]\n"
			  "[-firstclues <m>] [-uquick] [-unav <file w/ unavoidables>] [-x] [-dg] [-stats]\n\n"
           "Checks if there are any n clues that uniquely determine the given "
           "Sudoku grid.\n\n"
           "<gridfile> -- text file containing a completed 9x9 Sudoku grid\n"
           "<n> -- the n to search for, must be in the range from 10-30\n"
           "-pseudos -- also search for pseudo-(n-1)-puzzles (with two completions)\n"
           "-solve <percentage> -- percentage of puzzles to be solved, where the\n"
           "default value is 100 (exhaustive search)\n\n"
           "-uquick -- find only a small collection of unavoidable sets (but is quicker)\n"
           "-unav <file w/ unavoidables> -- file containing further unavoidable sets of\n"
           "the grid (see readme.txt for details)\n\n"
           "-random -- do a random search (useful for obtaining a reliable estimate of the\n"
           "total search time relatively quickly)\n"
           "-firstclues <m> -- allows to select the first <m> clues, thereby performing\n"
           "a partial search only (useful to split the search into several jobs)\n\n"
           "-x -- use Sudoku-X rules (Sudoku-X = diagonal Sudoku)\n"
           "-dg -- use Sudoku-DG rules (DG = disjoint groups)\n\n"
           "-stats -- print some statistics (currently only one item: time spent in solver)";
   }
