// This software including the source code and documentation (the "SOFTWARE")
// is copyrighted.  You are free to use the SOFTWARE for any non-commercial
// purpose whatsoever provided you do not remove this copyright notice.
// If you wish to use the SOFTWARE for commercial purposes, please contact us
// to obtain a license.  The SOFTWARE is provided "as is", with absolutely no
// warranty.  Use it at your own risk.
// Website: http://www.math.ie/checker.html

// unavoid
// finds unavoidable sets in a Sudoku grid and computes the MCN
// (MCN = max clique number, the maximum number of pairwise disjoint unavoidable sets)
// to learn the syntax, just run 'unavoid' with no parameters

#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>
#include <time.h>
#include <string>
#include <iostream>
#include "unavsets.h"
#include "gunav.h"
#include "unique.h"
#include "config.h"
#include "misc.h"
#include "mywin32.h"
#include "graph.h"

using namespace std;

void PrintSyntax();

int main( int argc, char const* argv[] )
   {
   if ( argc == 1 )
      {
      PrintSyntax();
      return -1;
      }

   EnableFindAllUnavoidables( true ); // by default, find *all* unavoidables
   bool bPrintSets = false;
   bool bMCNonly = false;
   char const* pszFileExtraUnav = NULL;
   for ( int i = 2;  i < argc;  i++ )
      {
      if ( 0 == strcmp( argv[ i ], "-unav" ) )
         {
         if ( ++i == argc )
            {
            // need one further argument
            printf( "Error: must specify name of file containing external unavoidables.\n" );            
            return -1;
            }
         else
            pszFileExtraUnav = argv[ i ];
         }
      else if ( 0 == strcmp( argv[ i ], "-mcn" ) )
         {
         if ( bPrintSets )
            {
            printf( "Error: cannot use both -p and -mcn simultaneously.\n" );          
            return -1;
            }
         bMCNonly = true;
         }
      else if ( 0 == strcmp( argv[ i ], "-p" ) )
         {
         if ( bMCNonly )
            {
            printf( "Error: cannot use both -p and -mcn simultaneously.\n" );          
            return -1;
            }
         bPrintSets = true;
         }
      else if ( 0 == strcmp( argv[ i ], "-x" ) )
         {
         if ( IsSudokuDG() )
            {
            printf( "Error: cannot use both -x and -dg simultaneously.\n" );           
            return -1;
            }
         EnableSudokuX();
         }
      else if ( 0 == strcmp( argv[ i ], "-dg" ) )
         {
         if ( IsSudokuX() )
            {
            printf( "Error: cannot use both -x and -dg simultaneously.\n" );           
            return -1;
            }
         EnableSudokuDG();
         }
      else if ( 0 == strcmp( argv[ i ], "-uquick" ) )
         {
         // find only a limited collection of unavoidables
         EnableFindAllUnavoidables( false );
         }
      else
         printf( "Ignoring unknown command-line argument %s.\n", argv[ i ] );
      }

   ifstream file( argv[ 1 ] );
   if ( ! file )
      {
      cout << "Unable to open file " << argv[ 1 ] << "." << endl;
      return -1;
      }

#ifdef _WIN32
   // assign lowest priority so that checker runs in the background and
   // does not slow down other processes
   SetPriorityClass( GetCurrentProcess(), IDLE_PRIORITY_CLASS );
#endif

   // initialise solver
   InitSolver();

   char szLine[ 100 ];
   int bFileEmpty = true;
   bool bIsThisTheFirstGrid = true;
   while ( file.getline( szLine, sizeof( szLine ) ) )
      {
      bFileEmpty = false;
      if ( strlen( szLine ) < 81 )
         {
         cout << "Erroneous line encountered (length < 81 ), skipping line." << endl;
         continue;
         }

      int A[ 9 ][ 9 ];
      bool bSkip = false;
      for ( int i = 0;  i < 81;  i++ )
         {
         if ( szLine[ i ] < '1'  ||  szLine[ i ] > '9' )
            {
            cout << "Illegal character encountered: '" << szLine[ i ] << "' (only digits 1-9 are allowed)."
                    " Skipping line." << endl;
            bSkip = true;
            continue;
            }
         A[ i / 9 ][ i % 9 ] = szLine[ i ] - '0';
         }
      if ( bSkip )
         continue;

      // print grid
      if ( ! bMCNonly )
         {
         for ( int y = 0;  y < 9;  y++ )
            {
            for ( int x = 0;  x < 9;  x++ )
               cout << A[ y ][ x ];
            cout << endl;
            }
         cout << endl;
         }

      vector< MySet > sets;
      int size4or6 = 0;
      int cSets = 0;
      if ( bIsThisTheFirstGrid )
         {
         // try to read the unavoidables from a file (which should exist if
         // unavoid or checker has been run on this grid before)
         // only for the first grid in a file
         std::string sFileUnav = GetUnavFilename( argv[ 1 ] );
         cSets = ReadAdditionalUnavoidables( sFileUnav.c_str(), sets );
         if ( cSets > 0 )
            {
            sort( sets.begin(), sets.end() );
            RemoveDuplicateSets( sets );
            TestSetsForUnavoidability( sets, szLine );
            cout << "Read " << ( int ) sets.size() << " unavoidable sets from " <<
                     sFileUnav.c_str() << " (will not compute\nunavoidable sets again, "
                     "please delete the file " << sFileUnav.c_str() << " to\ncompute "
                     "the unavoidable sets for this grid from scratch)." << endl << endl;
            // count the # of sets of size 4 or 6
            for ( vector< MySet >::iterator it = sets.begin();  it != sets.end();  it++ )
               if ( (*it).GetSize() == 4  ||  (*it).GetSize() == 6 )
                  size4or6++;
            }
         }

      if ( cSets <= 0 )
         {
         // no sets loaded from file, so do compute the unavoidable sets now
         // first those of size 4 and size 6
         FindUnavoidableSets( A, sets );

         // now compute the transpose of the matrix and find all unavoidable sets for the transpose
         int A_transpose[ 9 ][ 9 ];
         for ( int i = 0;  i < 9;  i++ )
            for ( int j = 0;  j < 9;  j++ )
               A_transpose[ i ][ j ] = A[ j ][ i ];
         // third parameter = true means swap rows and columns (to un-transpose the coordinates)
         FindUnavoidableSets( A_transpose, sets, true );
         // certain unavoidable sets of size 6 are found twice, remove them now
         sort( sets.begin(), sets.end() );
         RemoveDuplicateSets( sets );
         TestSetsForUnavoidability( sets, szLine );
         size4or6 = static_cast< int >( sets.size() );

         // now find (special) unavoidable sets using Guenter's code
         int B[ 82 ];
         for ( int i = 0;  i < 81;  i++ )
            B[ i + 1 ] = A[ i / 9 ][ i % 9 ];
         GFindUnavoidableSets( B, sets );

         if ( ShouldWeFindAllUnavoidables() )
            {
            // find unavoidables of sizes 8 to 12 (lengthy)
            cout << "Finding unavoidable sets (this takes about two minutes on "
                    "a 2 GHz computer).\n";
            FindUnavoidablesGeneral( A, sets );
            FindUnavoidablesGeneral( A_transpose, sets, true );
            cout << endl << endl;
            if ( bIsThisTheFirstGrid )
               {
               // save unavoidables to output file (for the first grid in the file only [in 
               // case there are several grids in a file])
               sort( sets.begin(), sets.end() );
               RemoveDuplicateSets( sets );
               TestSetsForUnavoidability( sets, szLine ); // crucial for Sudoku-X etc.
               SaveUnavoidablesToFile( GetUnavFilename( argv[ 1 ] ).c_str(), sets );
               }
            }
         }

      // read any additional unavoidable sets (from a file supplied by the user)
      if ( pszFileExtraUnav != NULL )
         {
         cSets = ReadAdditionalUnavoidables( pszFileExtraUnav, sets );
         if ( cSets > 0 )
            cout << "Read " << cSets << " additional unavoidable sets from " <<
                    argv[ 3 ] << endl;
         else if ( cSets == -1 )
            cout << "Unable to read external unavoidables from file " << pszFileExtraUnav << "." << endl;      
         }
      sort( sets.begin(), sets.end() );
      RemoveDuplicateSets( sets );
      // test if all the candidates found are actually unavoidable, remove those that aren't
      // (for ordinary/classical Sudoku, all sets found *should* be unavoidable, however
      // for variants like Sudoku-X, calling this function is absolutely crucial)
      TestSetsForUnavoidability( sets, szLine );

      if ( ! bMCNonly )
         {
         if ( bPrintSets )
            {
            // print all unavoidable sets
            cout << "Found " << ( int ) sets.size() << " unavoidable sets";
            cout << " (" << size4or6 << " of size 4 or 6):" << endl;
            for ( vector< MySet >::iterator it = sets.begin();  it != sets.end();  it++ )
               (*it).Print();
            cout << endl;
            }
         else
            {
            cout << "Found " << ( int ) sets.size() << " unavoidable sets";
            if ( ! IsSudokuX()  &&  ! IsSudokuDG() )
               cout << " (" << size4or6 << " of size 4 or 6)." << endl << endl;
            else
               cout << "." << endl << endl;
            }
         }

      // set up a graph whoses vertices are the unavoidable sets we have
      // two vertices are adjacent <=> the corresponding sets are disjoint
      // find a maximal clique in this graph
      vector< int > clique;
      Graph g( static_cast< int >( sets.size() ) );
      for ( int i = 0;  i < static_cast< int >( sets.size() );  i++ )
         for ( int j = i;  j < static_cast< int >( sets.size() );  j++ )
            if ( sets[ i ].Disjoint( sets[ j ] ) )
               g.AddEdge( i, j );
      // find a max clique in this graph
      g.FindMaxClique( &clique );
      sort( clique.begin(), clique.end() );

      if ( ! bMCNonly )
         {
         cout << "The maximum # of disjoint unavoidable sets (max clique number -- MCN) is " <<
                  static_cast< int >( clique.size() ) << "." << endl;
         cout << "One such maximal collection is:" << endl;
         int* anMemberOfDisj = new int[ sets.size() ];
         memset( anMemberOfDisj, 0, sizeof( int ) * sets.size() );
         //memset( abMemberOfDisj, sizeof( int ) * sets.size(), 0 );
         for ( vector< int >::iterator it = clique.begin();  it != clique.end();  it++ )
            {
            cout << "   ";
            sets[ *it ].Print();
            anMemberOfDisj[ *it ] = 1;
            }
         cout << endl << endl;
         }
      else
         {
         if ( ! bIsThisTheFirstGrid )
            cout << ", ";
         cout << static_cast< int >( clique.size() );
         }

      bIsThisTheFirstGrid = false;
      }

   if ( bFileEmpty )
      cout << "Error: empty input file encountered." << endl << endl;
   if ( bMCNonly )
      cout << endl;

   return 0;
   }


void PrintSyntax()
   {
   cout << "unavoid <gridfile> [-p] [-mcn] [-uquick] [-unav <file w/ unavoidables>]\n"
           "   [-x] [-dg]\n\n"
           "Finds unavoidable sets and computes the MCN for the given Sudoku grid.\n\n"
           "<gridfile> -- text file containing a completed 9x9 Sudoku grid\n\n"
           "-p -- print all unavoidable sets\n"
           "-mcn -- print MCN only (MCN = max clique number, the maximum number of\n"
           "  disjoint unavoidable sets), cannot be used in conjunction with -p\n\n"
           "-uquick -- find only a small collection of unavoidable sets (but is quicker)\n"
           "-unav <file w/ unavoidables> -- file containing further unavoidable sets of\n"
           "  the grid (see readme.txt for details)\n\n"
           "-x -- use Sudoku-X rules (Sudoku-X = diagonal Sudoku)\n"
           "-dg -- use Sudoku-DG rules (DG = disjoint groups)\n";
   }
